@@ -67,6 +67,16 @@
 		return errorstr;
 	}
 	#define strerror(x) print_last_windows_error()
+#else
+#ifdef HAVE_SELINUX
+	#include <selinux/selinux.h>
+	static char		*selinux_context = NULL;
+#endif
+	static uid_t		uid = 0;
+	static gid_t		gid = 0;
+	static char		*root_dir = NULL;
+	static bool		daemonize = false;
+	static FILE		*pid_file = NULL;
 #endif /* WIN32 */
 
 
@@ -88,10 +98,10 @@ int					tcp_port				= -1,		//	Port to send data to from the proxy
 					max_tunnels				= kMax_tunnels,	//	Default maximum number of tunnels to support at once
 					num_tunnels				= 0,			//	Current tunnel count
 					use_udp					= 0;			//	True if UDP should be used for transport (proxy runs on port 53)
-uint32_t			*seq_expiry_tbl			= 0;			//	Table indicating when a connection ID is allowable (used by proxy)
-char				*dest_proxy_addr		= 0,			//	Proxy's internet address (IP- or hostname)
-					*dest_tcp_addr			= 0,			//	Destination client wants data forwarded to (IP- or hostname)
-					*password				= 0,			//	Password (must be the same on proxy and client for authentication to succeed)
+uint32_t			*seq_expiry_tbl			= 0,			//	Table indicating when a connection ID is allowable (used by proxy)
+					given_proxy_ip			= 0,			//	Proxy's internet address
+					given_dst_ip			= 0;			//	Destination client wants data forwarded to
+char				*password				= 0,			//	Password (must be the same on proxy and client for authentication to succeed)
 					password_digest[kMD5_digest_size],		//	MD5 digest of password
 					*pcap_device			= 0;			//	Device to capture packets from
 
@@ -107,8 +117,14 @@ const char			*state_name[kNum_proto_types]	= { "start", "ack", "data", "close",
 
 //	Let the fun begin!
 int		main(int argc, char *argv[]) {
-	int			i, opt;
-	md5_state_t	state;
+	int				i, opt;
+	md5_state_t		state;
+	struct hostent	*host_ent;
+#ifndef WIN32
+	struct passwd	*pwnam;
+	struct group	*grnam;
+	pid_t			pid;
+#endif
 	
 	//	Seed random generator; it'll be used in combination with a timestamp
 	//	when generating authentication challenges.
@@ -153,13 +169,27 @@ int		main(int argc, char *argv[]) {
 		#ifndef WIN32
 		else if (strcmp(argv[i], "-syslog") == 0)
 			use_syslog		= !use_syslog;
+		else if (strcmp(argv[i], "-setuid") == 0)
+			opt	= kOpt_set_unpriv_user;
+		else if (strcmp(argv[i], "-setgid") == 0)
+			opt	= kOpt_set_unpriv_group;
+		else if (strcmp(argv[i], "-chroot") == 0)
+			opt	= kOpt_set_root_dir;
+		else if (strcmp(argv[i], "-setcon") == 0)
+			opt	= kOpt_set_selinux_context;
+		else if (strcmp(argv[i], "-daemon") == 0)
+			opt	= kOpt_daemonize;
 		#endif /* !WIN32 */
 		else if (strcmp(argv[i], "-udp") == 0)
 			use_udp			= 1;
 		else {
 			switch (opt) {
 				case kOpt_set_proxy_addr:
-					dest_proxy_addr			= argv[i];
+					if (NULL == (host_ent = gethostbyname(argv[i]))) {
+						pt_log(kLog_error, "Failed to look up %s as proxy address\n", argv[i]);
+						return 1;
+					}
+					given_proxy_ip = *(uint32_t*)host_ent->h_addr_list[0];
 					break;
 				case kOpt_set_password:
 					password				= argv[i];
@@ -168,12 +198,18 @@ int		main(int argc, char *argv[]) {
 					md5_init(&state);
 					md5_append(&state, (md5_byte_t*)password, strlen(password));
 					md5_finish(&state, (md5_byte_t*)password_digest);
+					//	Hide the password in process listing
+					memset(argv[i], ' ', strlen(argv[i]));
 					break;
 				case kOpt_set_tcp_port:
 					tcp_listen_port			= atoi(argv[i]);
 					break;
 				case kOpt_set_tcp_dest_addr:
-					dest_tcp_addr			= argv[i];
+					if (NULL == (host_ent = gethostbyname(argv[i]))) {
+						pt_log(kLog_error, "Failed to look up %s as destination address\n", argv[i]);
+						return 1;
+					}
+					given_dst_ip = *(uint32_t*)host_ent->h_addr_list[0];
 					break;
 				case kOpt_set_tcp_dest_port:
 					tcp_port				= atoi(argv[i]);
@@ -198,6 +234,42 @@ int		main(int argc, char *argv[]) {
 						pt_log(kLog_error, "Reverting log to standard out.\n");
 					}
 					break;
+				#ifndef WIN32
+				case kOpt_set_unpriv_user:
+					errno = 0;
+					if (NULL == (pwnam = getpwnam(argv[i]))) {
+						pt_log(kLog_error, "%s: %s\n", argv[i], errno ? strerror(errno) : "unknown user");
+						exit(1);
+					}
+					uid = pwnam->pw_uid;
+					if (!gid)
+						gid = pwnam->pw_gid;
+					break;
+				case kOpt_set_unpriv_group:
+					errno = 0;
+					if (NULL == (grnam = getgrnam(argv[i]))) {
+						pt_log(kLog_error, "%s: %s\n", argv[i], errno ? strerror(errno) : "unknown group");
+						exit(1);
+					}
+					gid = grnam->gr_gid;
+					break;
+				case kOpt_set_root_dir:
+					root_dir = strdup(argv[i]);
+					break;
+				case kOpt_set_selinux_context:
+				#ifdef HAVE_SELINUX
+					selinux_context = strdup(argv[i]);
+				#else
+					pt_log(kLog_error, "Sorry: SELinux support missing, please recompile with libselinux.\n");
+					return 1;
+				#endif
+					break;
+				case kOpt_daemonize:
+					daemonize = true;
+					if (NULL == (pid_file = fopen(argv[i], "w")))
+						pt_log(kLog_error, "%s: %s\n", argv[i], strerror(errno));
+					break;
+				#endif /* !WIN32 */
 				case kOpt_undefined:
 					usage(argv[0]);
 					return 1;
@@ -217,6 +289,8 @@ int		main(int argc, char *argv[]) {
 	pt_log(kLog_info, "(c) 2004-2009 Daniel Stoedle, <daniels@cs.uit.no>\n");
 	#ifdef WIN32
 	pt_log(kLog_info, "Windows version by Mike Miller, <mike@mikeage.net>\n");
+	#else
+	pt_log(kLog_info, "Security features by Sebastien Raveau, <sebastien.raveau@epita.fr>\n");
 	#endif
 	pt_log(kLog_info, "%s.\n", (mode == kMode_forward ? "Relaying packets from incoming TCP streams" : "Forwarding incoming ping packets over TCP"));
 	if (use_udp)
@@ -232,6 +306,38 @@ int		main(int argc, char *argv[]) {
 		}		
 		openlog("ptunnel", LOG_PID, LOG_USER);
 	}
+	if (NULL != root_dir) {
+		pt_log(kLog_info, "Restricting file access to %s\n", root_dir);
+		if (-1 == chdir(root_dir) || -1 == chroot(root_dir)) {
+			pt_log(kLog_error, "%s: %s\n", root_dir, strerror(errno));
+			exit(1);
+		}
+	}
+	if (daemonize) {
+		pt_log(kLog_info, "Going to the background.\n");
+		if (0 < (pid = fork()))
+			exit(0);
+		if (0 > pid)
+			pt_log(kLog_error, "fork: %s\n", strerror(errno));
+		else
+			if (-1 == setsid())
+				pt_log(kLog_error, "setsid: %s\n", strerror(errno));
+			else {
+				if (0 < (pid = fork()))
+					exit(0);
+				if (0 > pid)
+					pt_log(kLog_error, "fork: %s\n", strerror(errno));
+				else {
+					if (NULL != pid_file) {
+						fprintf(pid_file, "%d\n", getpid());
+						fclose(pid_file);
+					}
+					freopen("/dev/null", "r", stdin);
+					freopen("/dev/null", "w", stdout);
+					freopen("/dev/null", "w", stderr);
+				}
+			}
+	}
 #endif /* !WIN32 */
 
 #ifdef WIN32
@@ -257,7 +363,7 @@ int		main(int argc, char *argv[]) {
 	
 	//	Check mode, validate arguments and start either client or proxy.
 	if (mode == kMode_forward) {
-		if (!dest_proxy_addr || !dest_tcp_addr || !tcp_port || !tcp_listen_port) {
+		if (!given_proxy_ip || !given_dst_ip || !tcp_port || !tcp_listen_port) {
 			printf("One of the options are missing or invalid.\n");
 			usage(argv[0]);
 			return -1;
@@ -273,6 +379,13 @@ int		main(int argc, char *argv[]) {
 
 #ifdef WIN32
 	WSACleanup();
+#else
+	if (NULL != root_dir)
+		free(root_dir);
+	#ifdef HAVE_SELINUX
+	if (NULL != selinux_context)
+		free(selinux_context);
+	#endif
 #endif /* WIN32 */
 
 	pt_log(kLog_info, "ptunnel is exiting.\n");
@@ -282,30 +395,43 @@ int		main(int argc, char *argv[]) {
 
 void		usage(char *exec_name) {
 	printf("ptunnel v %d.%.2d.\n", kMajor_version, kMinor_version);
-	printf("Usage:   %s -p <addr> -lp <port> -da <dest_addr> -dp <dest_port> [-m max_tunnels] [-x password] [-v verbosity] [-f logfile]\n", exec_name);
-	printf("         %s [-m max_threads] [-x password] [-v verbosity] [-c <device>]\n", exec_name);
+	printf("Usage:   %s -p <addr> -lp <port> -da <dest_addr> -dp <dest_port> [-m max_tunnels] [-v verbosity] [-f logfile]\n", exec_name);
+	printf("         %s [-m max_threads] [-v verbosity] [-c <device>]\n", exec_name);
 	printf("     -p: Set address of peer running packet forwarder. This causes\n");
 	printf("         ptunnel to operate in forwarding mode - the absence of this\n");
 	printf("         option causes ptunnel to operate in proxy mode.\n");
 	printf("    -lp: Set TCP listening port (only used when operating in forward mode)\n");
-	printf("    -da: Set remote proxy destination address\n");
-	printf("    -dp: Set remote proxy destionation port\n");
+	printf("    -da: Set remote proxy destination address if client\n");
+	printf("         Restrict to only this destination address if server\n");
+	printf("    -dp: Set remote proxy destionation port if client\n");
+	printf("         Restrict to only this destination port if server\n");
 	printf("     -m: Set maximum number of concurrent tunnels\n");
-	printf("     -x: Set password (must be same on client and proxy)\n");
-	printf("     -u: Run proxy in unprivileged mode. This causes the proxy to forward\n");
-	printf("         packets using standard echo requests, instead of crafting custom echo replies.\n");
-	printf("         Unprivileged mode will only work on some systems, and is in general less reliable\n");
-	printf("         than running in privileged mode.\n");
 	printf("     -v: Verbosity level (-1 to 4, where -1 is no output, and 4 is all output)\n");
 	printf("     -c: Enable libpcap on the given device.\n");
 	printf("     -f: Specify a file to log to, rather than printing to standard out.\n");
 	printf("     -s: Client only. Enables continuous output of statistics (packet loss, etc.)\n");
 	#ifndef WIN32
+	printf("-daemon: Run in background, the PID will be written in the file supplied as argument\n");
 	printf("-syslog: Output debug to syslog instead of standard out.\n");
 	#endif /* !WIN32 */
-
 	printf("   -udp: Toggle use of UDP instead of ICMP. Proxy will listen on port 53 (must be root).\n\n");
-	printf("Starting the proxy (needs to run as root):\n");
+
+	printf("Security features:  [-x password] [-u] [-setuid user] [-setgid group] [-chroot dir]\n");
+	printf("     -x: Set password (must be same on client and proxy)\n");
+	printf("     -u: Run proxy in unprivileged mode. This causes the proxy to forward\n");
+	printf("         packets using standard echo requests, instead of crafting custom echo replies.\n");
+	printf("         Unprivileged mode will only work on some systems, and is in general less reliable\n");
+	printf("         than running in privileged mode.\n");
+	#ifndef WIN32
+	printf("         Please consider combining the following three options instead:\n");
+	printf("-setuid: When started in privileged mode, drop down to user's rights as soon as possible\n");
+	printf("-setgid: When started in privileged mode, drop down to group's rights as soon as possible\n");
+	printf("-chroot: When started in privileged mode, restrict file access to the specified directory\n");
+	printf("-setcon: Set SELinux context when all there is left to do are network I/O operations\n");
+	printf("         To combine with -chroot you will have to `mount --bind /proc /chrootdir/proc`\n");
+	#endif /* !WIN32 */
+
+	printf("\nStarting the proxy (needs to run as root):\n");
 	printf(" [root #] %s\n", exec_name);
 	printf("Starting a client (also needs root):\n");
 	printf(" [root #] %s -p proxy.pingtunnel.com -lp 8000 -da login.domain.com -dp 22 -c eth0\n", exec_name);
@@ -326,9 +452,7 @@ void		pt_forwarder(void) {
 	struct sockaddr_in	addr, dest_addr;
 	socklen_t			addr_len;
 	pthread_t			pid;
-	uint32_t			dst_ip, proxy_ip;
 	uint16_t			rand_id;
-	struct hostent		*host_ent;
 	
 	pt_log(kLog_debug, "Starting forwarder..\n");
 	//	Open our listening socket
@@ -348,22 +472,6 @@ void		pt_forwarder(void) {
 		return;
 	}
 	server_sock		= sock;
-	//	Look up IP of final destination (ie, where the proxy should send the packets)
-	host_ent		= gethostbyname(dest_tcp_addr);
-	if (!host_ent) {
-		pt_log(kLog_error, "Failed to look up destination address\n");
-		close(sock);
-		return;
-	}
-	dst_ip			= *(uint32_t*)host_ent->h_addr_list[0];
-	//	Look up proxy's IP
-	host_ent		= gethostbyname(dest_proxy_addr);
-	if (!host_ent) {
-		pt_log(kLog_error, "Failed to look up proxy IP address\n");
-		close(sock);
-		return;
-	}
-	proxy_ip		= *(uint32_t*)host_ent->h_addr_list[0];
 	//	Fill out address structure
 	memset(&dest_addr, 0, sizeof(struct sockaddr_in));
 	dest_addr.sin_family			= AF_INET;
@@ -371,8 +479,8 @@ void		pt_forwarder(void) {
 		dest_addr.sin_port			= htons(kDNS_port /* dns port.. */);
 	else
 		dest_addr.sin_port			= 0;
-	dest_addr.sin_addr.s_addr		= proxy_ip;
-	pt_log(kLog_verbose, "Proxy IP address: %s\n", inet_ntoa(*((struct in_addr*)&proxy_ip)));
+	dest_addr.sin_addr.s_addr		= given_proxy_ip;
+	pt_log(kLog_verbose, "Proxy IP address: %s\n", inet_ntoa(*((struct in_addr*)&given_proxy_ip)));
 	
 	listen(server_sock, 10);
 	while (1) {
@@ -405,7 +513,7 @@ void		pt_forwarder(void) {
 			}
 			addr	= dest_addr;
 			rand_id	= (uint16_t)rand();
-			create_and_insert_proxy_desc(rand_id, rand_id, new_sock, &addr, dst_ip, tcp_port, kProxy_start, kUser_flag);
+			create_and_insert_proxy_desc(rand_id, rand_id, new_sock, &addr, given_dst_ip, tcp_port, kProxy_start, kUser_flag);
 			pthread_mutex_unlock(&num_threads_lock);
 		}
 	}
@@ -549,6 +657,24 @@ void*		pt_proxy(void *args) {
 	
 	//	Start forwarding :)
 	pt_log(kLog_info, "Ping proxy is listening in %s mode.\n", (unprivileged ? "unprivileged" : "privileged"));
+
+	#ifndef WIN32
+	#ifdef HAVE_SELINUX
+	if (uid || gid || selinux_context)
+	#else
+	if (uid || gid)
+	#endif
+		pt_log(kLog_info, "Dropping privileges now.\n");
+	if (gid && -1 == setgid(gid))
+		pt_log(kLog_error, "setgid(%d): %s\n", gid, strerror(errno));
+	if (uid && -1 == setuid(uid))
+		pt_log(kLog_error, "setuid(%d): %s\n", uid, strerror(errno));
+	#ifdef HAVE_SELINUX
+	if (NULL != selinux_context && -1 == setcon(selinux_context))
+		pt_log(kLog_error, "setcon(%s) failed: %s\n", selinux_context, strerror(errno));
+	#endif
+	#endif
+
 	while (1) {
 		FD_ZERO(&set);
 		FD_SET(fwd_sock, &set);
@@ -784,7 +910,7 @@ void		handle_packet(char *buf, int bytes, int is_pcap, struct sockaddr_in *addr,
 	struct timeval		tt;
 	
 	if (bytes < sizeof(icmp_echo_packet_t)+sizeof(ping_tunnel_pkt_t))
-		pt_log(kLog_verbose, "Skipping this packet - too short.\n");
+		pt_log(kLog_verbose, "Skipping this packet - too short. Expect: %d+%d = %d ; Got: %d\n", sizeof(icmp_echo_packet_t), sizeof(ping_tunnel_pkt_t), sizeof(icmp_echo_packet_t)+sizeof(ping_tunnel_pkt_t), bytes);
 	else {
 		if (use_udp) {
 			ip_pkt		= 0;
@@ -845,6 +971,10 @@ void		handle_packet(char *buf, int bytes, int is_pcap, struct sockaddr_in *addr,
 							return;
 						}
 						pt_log(kLog_info, "Starting new session to %s:%d with ID %d\n", inet_ntoa(*(struct in_addr*)&pt_pkt->dst_ip), ntohl(pt_pkt->dst_port), pt_pkt->id_no);
+						if ((given_dst_ip && given_dst_ip != pt_pkt->dst_ip) || (-1 != tcp_port && tcp_port != ntohl(pt_pkt->dst_port))) {
+							pt_log(kLog_info, "Destination administratively prohibited!\n");
+							return;
+						}
 						if (password)
 							init_state	= kProto_authenticate;
 						else
